/*********************************************************************
* MATHERR.C - This program demonstrates how to handle exceptions in  *
* the run-time math library functions by defining your own matherr() *
* function. In addition, the program contains new and improved       *
* versions of the probability and statistics functions first that we *
* first presented last month.                                        *
*                                                                    *
* To compile: CL MATHERR.C /link /NOE                                *
*********************************************************************/
#include <stdio.h>      /* stderr                                   */
#include <stdlib.h>     /* printf()                                 */
#include <math.h>       /* pow(), EDOM, ERANGE, DOMAIN, SING, ...   */
#include <float.h>      /* MAX_DBL                                  */

/*----PROTOTYPES----------------------------------------------------*/
int matherr_A( struct exception *except );
int matherr_B( struct exception *except );
char *TranslateMathError( int MathErrorNum );

double Factorial( unsigned int N );
double Permutation( unsigned int N, unsigned int R );
double Combination( int N, int R );
double Binomial( int N, int X, double P );
double Hypergeometric( int N, int K, int S, int X );
		
/*----GLOBAL VARIABLES----------------------------------------------*/
int MathError;
struct exception MathExcept;
int (*CurrMathErr)(struct exception *except);

/********************************************************************/
int main ( void )
    {
    int i;
    double d;

    for ( i=0; i<3; i++ )
        {
        printf( "\n**********PASS NUMBER %d**********\n", i );
        switch ( i )
            {
            case 0 : CurrMathErr = NULL;
                     break;
            case 1 : CurrMathErr = matherr_A;
                     break;
            case 2 : CurrMathErr = matherr_B;
                     break;
            }

        printf( "TEST1: " );
        MathError = 0;
        d = Factorial( 171 );
        if ( MathError )
            printf( "Error computing factorial: %s\n",
                    TranslateMathError( MathError ) );

        printf( "TEST2: " );
        MathError = 0;
        d = Combination( 5000, 500 );
        if ( MathError )
            printf( "Error computing combination: %s\n",
                    TranslateMathError( MathError ) );

        printf( "TEST3: " );
        MathError = 0;
        d = Binomial( 42, 0, 0.0 );
        if ( MathError )
            printf( "Error computing binomial probability: %s\n",
                    TranslateMathError( MathError ) );

        printf( "TEST4: " );
        MathError = 0;
        d = Hypergeometric( 5000, 2000, 500, 30 );
        if ( MathError )
            printf( "Error computing hypergeometric probability: "
                    "%s\n", TranslateMathError( MathError ) );
        }
    }

/********************************************************************/
int matherr( struct exception *except )
    {
    if ( CurrMathErr != NULL )
        return( (*CurrMathErr)(except) );
    else
        return( 0 );
    }

/********************************************************************/
int matherr_A( struct exception *except )
{
  char errmess[80];

  sprintf( errmess, "MATHERR: %s in %s function\n",
           TranslateMathError(except->type), except->name );
  fprintf( stderr, errmess );
  return( 1 );
}


/********************************************************************/
int matherr_B( struct exception *except )
    {
    MathError = except->type;
    if ( MathError == DOMAIN )
        errno = EDOM;
    else
        errno = ERANGE;
    return( 1 );
    }

/********************************************************************/
char *TranslateMathError( int MathErrorNum )
    {
    switch ( MathErrorNum )
        {
        case DOMAIN    : return( "argument domain error" );
        case SING      : return( "argument singularity" );
        case OVERFLOW  : return( "overflow range error" );
        case UNDERFLOW : return( "underflow range error" );
        case PLOSS	    : return( "partial loss of significance" );
        case TLOSS	    : return( "total loss of significance" );
        }
    return( "" );   /* Just in case MathError == 0 or invalid value */
    }

/********************************************************************/
double Factorial( unsigned int N )
    {
    int i;
    double Rval;

    if ( N > 170 )
        {
        errno = ERANGE;
        MathError = OVERFLOW;
        return( DBL_MAX );
        }
    
    Rval = 1.0;
    for ( i=2; i<=N; i++ )
       Rval *= i;

    return( Rval );
    }

/********************************************************************/
double Permutation( unsigned int N, unsigned int R )
    {
    int i;
    long double Rval = R ? 1.0 : 0.0;

    for ( i=N; i>N-R; i-- )
        {
        Rval *= i;
        if ( Rval > DBL_MAX )
            {
            errno = ERANGE;
            MathError = OVERFLOW;
            return( DBL_MAX );
            }
        }

    return( Rval );
    }

/********************************************************************/
double Combination( int N, int R )
    {
    unsigned int i, j;
    long double Rval = R ? 1.0 : 0.0;

    for ( i=N, j=R; i>N-R; i--, j-- )
        {
        Rval *= (double)i / (double)j;
        if ( Rval > DBL_MAX )
            {
            errno = ERANGE;
            MathError = OVERFLOW;
            return( DBL_MAX );
            }
        }

    return( Rval );
    }

/********************************************************************/
double Binomial( int N, int X, double P )
    {
    return( Combination( N, X ) 
            * pow( P, (double)X ) 
            * pow( 1.0 - P, (double)(N - X) ) );
    }

/********************************************************************/
double Hypergeometric( int N, int K, int S, int X )
    {
/*----Uncomment this code to avoid floating-point divide by zero----*/
//  if ( S == 0 )
//      {
//      errno = ERANGE;
//      MathError = SING;
//      return( 0.0 );
//      }

    return( Combination( K, X ) * Combination( N-K, S-X )
            / Combination( N, S ) );
    }
