/*********************************************************************
* RLE.C  -  This program performs Run Length Encoding (RLE) on an    *
* input file. The input and output file names are provided on the    *
* command line. Runs of matching characters 4 or more in length are  *
* compressed into the 3 byte following format:                       *
*                                                                    *
*     (char) (Repeat Identifier) (count)                             *
*                                                                    *
* To compile: CL RLE.C                                               *
* To run : RLE <infile> <outfile>                                    *
*                                                                    *
* Author  : Gary Conway                                              *
* Last Revised : 7-15-91 by David Reid                               *
*********************************************************************/

#include <stdio.h>             /* printf(), fopen(), fgetc(), ...   */
#include <stdlib.h>            /* exit(), EXIT_FAILURE              */

int Compressor( void );        /* Prototype for compressor function */

/*----Program constants---------------------------------------------*/
#define Repeat      0x90       /* RLE Repeat sequence identifier    */
#define MAXCOUNT    257        /* Max repeat sequence length        */

/*----RLE state-machine values--------------------------------------*/
#define FirstTime   0          /* don't consider previous input     */
#define LoneChar    1          /* lastchar sent by itself           */
#define SmallRun    2          /* outputting a short run of chars   */
#define SendNewChar 3          /* run finished, send new char next  */
#define SendRLECnt  4          /* send RLE count next               */

/*----Global variables----------------------------------------------*/
FILE *infile, *outfile;        /* Input and output files            */
static unsigned char State;    /* The current RLE state             */

/*********************************************************************
* main - Main body of program                                        *
*********************************************************************/
void main( int argc, char **argv )
    {
    int Ch;                    /* A character variable              */

    /*----Check for exactly two command line arguments--------------*/
    if ( argc != 3 )
        {
        puts( "SYNTAX: rle <infile> <outfile>" );
        exit( EXIT_FAILURE );
        }

    /*----Open the input file---------------------------------------*/
    if ( (infile = fopen( argv[1], "rb" )) == NULL )
        {
        printf( "Unable to open input file: %s\n", argv[1] );
        exit( EXIT_FAILURE );
        }

    /*----See if output file already exists-------------------------*/
    if ( (outfile = fopen( argv[2], "rb" )) != NULL )
        {
        printf( "Output file %s already exists", argv[2] );
        fclose( infile );
        fclose( outfile );
        exit( EXIT_FAILURE );
        }

    /*----Open the output file--------------------------------------*/
    if ( (outfile = fopen( argv[2], "wb" )) == NULL )
        {
        printf( "Unable to open output file: %s\n", argv[2] );
        fclose( infile );
        exit( EXIT_FAILURE );
        }

    /*----Run the compressor until end-of-file encountered----------*/
    State = FirstTime;
    while ( ( Ch = Compressor() ) != EOF )
        fputc( Ch, outfile );

    /*----Close input and output files------------------------------*/
    fclose( infile );
    fclose( outfile );
    }

/*********************************************************************
* The Run-Length-Encoding state-machine                              *
*********************************************************************/
int Compressor( void )
    {
    static int C;              /* current character for processing  */
    static int PrevC;          /* previous character                */
    static int SmallCount;     /* counter for < min length runs     */
    static int Count;          /* repeat sequence counter           */

    switch( State )
        {
        /*----Start new sequence------------------------------------*/
        case FirstTime:
            State = LoneChar;
            SmallCount = 0;
            PrevC = fgetc( infile );
            return( PrevC );

        /*----Look for run matching last character------------------*/
        case LoneChar:
            switch( PrevC )
                {
                case Repeat:   /* Last char was 0x90, output 0x00   */
                    State = FirstTime;
                    return( 0 );

                default:       /* Find runs and small runs          */
                    for ( Count = 0; /* Count up to MAXCOUNT repeats*/
                          PrevC == (C = fgetc(infile)) &&
                          Count < MAXCOUNT;
                          Count++ ) ;

                    /*----handle runs of 1,2,3, and 4-or-more chars-*/
                    switch( Count )
                    {
                        case 0:  /* single character -- no repeats  */
                            return( PrevC = C );
                        case 1:  /* character repeated once         */
                        case 2:  /* character repeated twice        */
                            State = SmallRun;
                            SmallCount = Count;
                            return( PrevC );
                        default: /* character repeated > 2 times    */
                            State = SendRLECnt;
                            return( Repeat );
                    }
                }

        /*----Output next byte from small run-----------------------*/
        case SmallRun:
            SmallCount--;
            if ( SmallCount == 0 ) /* finished small run*/
                {
                State = LoneChar;
                return( PrevC = C );
                }
            else                   /* continue small run*/
                return( PrevC );

        /*----Output run-length byte--------------------------------*/
        case SendRLECnt:
            State = SendNewChar;
            return( Count-2 );

        /*----Output new byte and reset to LoneChar state-----------*/
        case SendNewChar:
            State = LoneChar;
            return( PrevC = C );
        }
    }
